#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "NestedTempfile",
    "TempDir",
]

import os
import shutil
import tempfile
from typing import Optional

import fixtures


class TempDir(fixtures.Fixture):
    """Create a temporary directory.

    :ivar path: The path of the temporary directory.
    """

    path: str
    rootdir: Optional[str]

    def __init__(self, rootdir: Optional[str] = None) -> None:
        """Create a TempDir.

        :param rootdir: If supplied force the temporary directory to be a
            child of rootdir.
        """
        self.rootdir = rootdir

    def _setUp(self) -> None:
        self.path = tempfile.mkdtemp(dir=self.rootdir)
        self.addCleanup(shutil.rmtree, self.path, ignore_errors=True)

    def join(self, *children: str) -> str:
        """Return an absolute path, given one relative to this ``TempDir``.

        WARNING: This does not do any checking of ``children`` to make sure
        they aren't walking up the tree using path segments like '..' or
        '/usr'.  Use at your own risk.
        """
        return os.path.abspath(os.path.join(self.path, *children))


class NestedTempfile(fixtures.Fixture):
    """Nest all temporary files and directories inside another directory.

    This temporarily monkey-patches the default location that the `tempfile`
    package creates temporary files and directories in to be a new temporary
    directory. This new temporary directory is removed when the fixture is torn
    down.
    """

    def _setUp(self) -> None:
        tempdir_fixture = self.useFixture(TempDir())
        tempdir = tempdir_fixture.path
        patch = fixtures.MonkeyPatch("tempfile.tempdir", tempdir)
        self.useFixture(patch)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, 2011, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "EnvironmentVariable",
    "EnvironmentVariableFixture",
]

import os
from typing import Optional

from fixtures import Fixture


class EnvironmentVariable(Fixture):
    """Isolate a specific environment variable."""

    varname: str
    newvalue: Optional[str]

    def __init__(self, varname: str, newvalue: Optional[str] = None) -> None:
        """Create an EnvironmentVariable fixture.

        :param varname: the name of the variable to isolate.
        :param newvalue: A value to set the variable to. If None, the variable
            will be deleted.

        During setup the variable will be deleted or assigned the requested
        value, and this will be restored in cleanUp.
        """
        super(EnvironmentVariable, self).__init__()
        self.varname = varname
        self.newvalue = newvalue

    def _setUp(self) -> None:
        varname = self.varname
        orig_value = os.environ.get(varname)
        if orig_value is not None:
            self.addCleanup(os.environ.__setitem__, varname, orig_value)
            del os.environ[varname]
        else:
            self.addCleanup(os.environ.pop, varname, "")
        if self.newvalue is not None:
            os.environ[varname] = self.newvalue
        else:
            os.environ.pop(varname, "")


EnvironmentVariableFixture = EnvironmentVariable



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "TestWithFixtures",
]

import unittest
from typing import TypeVar

from fixtures.fixture import Fixture
import fixtures.fixture

gather_details = fixtures.fixture.gather_details

T = TypeVar("T", bound=Fixture)


class TestWithFixtures(unittest.TestCase):
    """A TestCase with a helper function to use fixtures.

    Normally used as a mix-in class to add useFixture.

    Note that test classes such as testtools.TestCase which already have a
    ``useFixture`` method do not need this mixed in.
    """

    def useFixture(self, fixture: T) -> T:
        """Use fixture in a test case.

        The fixture will be setUp, and self.addCleanup(fixture.cleanUp) called.

        :param fixture: The fixture to use.
        :return: The fixture, after setting it up and scheduling a cleanup for
           it.
        """
        use_details = (
            gather_details is not None and getattr(self, "addDetail", None) is not None
        )
        try:
            fixture.setUp()
        except:
            if gather_details is not None and use_details:
                # Capture the details now, in case the fixture goes away.
                get_details = getattr(self, "getDetails", None)
                if get_details is not None:
                    gather_details(fixture.getDetails(), get_details())
            raise
        else:
            self.addCleanup(fixture.cleanUp)
            if gather_details is not None and use_details:
                # Capture the details from the fixture during test teardown;
                # this will evaluate the details before tearing down the
                # fixture.
                def cleanup_details() -> None:
                    if gather_details is not None:
                        get_details = getattr(self, "getDetails", None)
                        if get_details is not None:
                            gather_details(fixture.getDetails(), get_details())

                self.addCleanup(cleanup_details)
            return fixture



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Canonical Ltd.
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "TempHomeDir",
]

from fixtures._fixtures.tempdir import TempDir
from fixtures._fixtures.environ import EnvironmentVariable


class TempHomeDir(TempDir):
    """Create a temporary directory and set it as $HOME

    :ivar path: the path of the temporary directory.
    """

    def _setUp(self) -> None:
        super(TempHomeDir, self)._setUp()
        self.useFixture(EnvironmentVariable("HOME", self.path))



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "PythonPackage",
]

import os.path
from typing import List, Tuple

from fixtures import Fixture
from fixtures._fixtures.tempdir import TempDir


class PythonPackage(Fixture):
    """Create a temporary Python package.

    :ivar base: The path of the directory containing the module. E.g. for a
        module 'foo', the path base + '/foo/__init__.py' would be the file path
        for the module.
    """

    def __init__(
        self, packagename: str, modulelist: List[Tuple[str, bytes]], init: bool = True
    ) -> None:
        """Create a PythonPackage.

        :param packagename: The name of the package to create - e.g.
            'toplevel.subpackage.'
        :param modulelist: List of modules to include in the package.
            Each module should be a tuple with the filename and content it
            should have.
        :param init: If false, do not create a missing __init__.py. When
            True, if modulelist does not include an __init__.py, an empty
            one is created.
        """
        self.packagename = packagename
        self.modulelist = modulelist
        self.init = init

    def _setUp(self) -> None:
        self.base = self.useFixture(TempDir()).path
        base = self.base
        root = os.path.join(base, self.packagename)
        os.mkdir(root)
        init_seen = not self.init
        for modulename, contents in self.modulelist:
            stream = open(os.path.join(root, modulename), "wb")
            try:
                stream.write(contents)
            finally:
                stream.close()
            if modulename == "__init__.py":
                init_seen = True
        if not init_seen:
            open(os.path.join(root, "__init__.py"), "wb").close()



# Copyright 2010 United States Government as represented by the
# Administrator of the National Aeronautics and Space Administration.
# Copyright 2013 Hewlett-Packard Development Company, L.P.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import unittest.mock as mock
from typing import Any, Callable

import fixtures


class _Base(fixtures.Fixture):
    _get_p: Callable[[], Any]  # Returns mock._patch or similar
    mock: mock.MagicMock

    def _setUp(self) -> None:
        _p = self._get_p()
        self.addCleanup(_p.stop)
        self.mock = _p.start()


class MockPatchObject(_Base):
    """Deal with code around mock."""

    def __init__(self, obj: Any, attr: str, new: Any = None, **kwargs: Any) -> None:
        super(MockPatchObject, self).__init__()
        if new is None:
            new = mock.DEFAULT
        self._get_p: Callable[[], Any] = lambda: mock.patch.object(
            obj, attr, new, **kwargs
        )


class MockPatch(_Base):
    """Deal with code around mock.patch."""

    def __init__(self, obj: str, new: Any = None, **kwargs: Any) -> None:
        super(MockPatch, self).__init__()
        if new is None:
            new = mock.DEFAULT
        self._get_p: Callable[[], Any] = lambda: mock.patch(obj, new, **kwargs)


class _MockPatchMultipleMeta(type):
    """Arrange for lazy loading of MockPatchMultiple.DEFAULT."""

    # For strict backward compatibility, ensure that DEFAULT also works as
    # an instance property.
    def __new__(
        cls,
        name: str,
        bases: tuple[type, ...],
        namespace: dict[str, Any],
        **kwargs: Any,
    ) -> type:
        namespace["DEFAULT"] = cls.DEFAULT
        return super().__new__(cls, name, bases, namespace, **kwargs)

    # Default value to trigger a MagicMock to be created for a named
    # attribute.
    @property
    def DEFAULT(self) -> Any:
        return mock.DEFAULT


class MockPatchMultiple(_Base, metaclass=_MockPatchMultipleMeta):
    """Deal with code around mock.patch.multiple."""

    def __init__(self, obj: Any, **kwargs: Any) -> None:
        """Initialize the mocks

        Pass name=value to replace obj.name with value.

        Pass name=Multiple.DEFAULT to replace obj.name with a
        MagicMock instance.

        :param obj: Object or name containing values being mocked.
        :type obj: str or object
        :param kwargs: names and values of attributes of obj to be mocked.

        """
        super(MockPatchMultiple, self).__init__()
        self._get_p: Callable[[], Any] = lambda: mock.patch.multiple(obj, **kwargs)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, 2011, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.


"""Fixtures provides a sensible contract for reusable test fixtures.

It also provides glue for using these in common test runners and acts as a
common repository for widely used Fixture classes.

See the README for a manual, and the docstrings on individual functions and
methods for details.

Most users will want to look at TestWithFixtures and Fixture, to start with.
"""

from fixtures._version import __version__

__all__ = [
    "ByteStream",
    "CompoundFixture",
    "DetailStream",
    "EnvironmentVariable",
    "EnvironmentVariableFixture",
    "FakeLogger",
    "FakePopen",
    "Fixture",
    "FunctionFixture",
    "LogHandler",
    "LoggerFixture",
    "MethodFixture",
    "MockPatch",
    "MockPatchMultiple",
    "MockPatchObject",
    "MonkeyPatch",
    "MultipleExceptions",
    "NestedTempfile",
    "PackagePathEntry",
    "PopenFixture",
    "PythonPackage",
    "PythonPathEntry",
    "SetupError",
    "StringStream",
    "TempDir",
    "TempHomeDir",
    "TestWithFixtures",
    "Timeout",
    "TimeoutException",
    "WarningsCapture",
    "WarningsFilter",
    "__version__",
]


from fixtures.fixture import (  # noqa: E402
    CompoundFixture,
    Fixture,
    FunctionFixture,
    MethodFixture,
    MultipleExceptions,
    SetupError,
)
from fixtures._fixtures import (  # noqa: E402
    ByteStream,
    DetailStream,
    EnvironmentVariable,
    EnvironmentVariableFixture,
    FakeLogger,
    FakePopen,
    LoggerFixture,
    LogHandler,
    MockPatch,
    MockPatchMultiple,
    MockPatchObject,
    MonkeyPatch,
    NestedTempfile,
    PackagePathEntry,
    PopenFixture,
    PythonPackage,
    PythonPathEntry,
    StringStream,
    TempDir,
    TempHomeDir,
    Timeout,
    TimeoutException,
    WarningsCapture,
    WarningsFilter,
)
from fixtures.testcase import TestWithFixtures  # noqa: E402



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2010, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "CompoundFixture",
    "Fixture",
    "FunctionFixture",
    "MethodFixture",
    "MultipleExceptions",
    "SetupError",
]

import itertools
import sys
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    Literal,
    Optional,
    Tuple,
    TypeVar,
    TYPE_CHECKING,
    Union,
)

from fixtures.callmany import (
    CallMany,
)

# Deprecated, imported for compatibility.
import fixtures.callmany

if TYPE_CHECKING:
    from types import TracebackType

T = TypeVar("T", bound="Fixture")

MultipleExceptions = fixtures.callmany.MultipleExceptions  # type: ignore[attr-defined]

gather_details: Union[Callable[[Dict[str, Any], Dict[str, Any]], None], None]
try:
    from testtools.testcase import gather_details as _gather_details

    gather_details = _gather_details
except ImportError:
    gather_details = None


# This would be better in testtools (or a common library)
def combine_details(
    source_details: Dict[str, Any], target_details: Dict[str, Any]
) -> None:
    """Add every value from source to target deduping common keys."""
    for name, content_object in source_details.items():
        new_name = name
        disambiguator = itertools.count(1)
        while new_name in target_details:
            new_name = "%s-%d" % (name, next(disambiguator))
        name = new_name
        target_details[name] = content_object


class SetupError(Exception):
    """Setup failed.

    args[0] will be a details dict.
    """


class Fixture(object):
    _cleanups: Optional[CallMany]
    _details: Optional[Dict[str, Any]]
    _detail_sources: Optional[List["Fixture"]]
    """A Fixture representing some state or resource.

    Often used in tests, a Fixture must be setUp before using it, and cleanUp
    called after it is finished with (because many Fixture classes have
    external resources such as temporary directories).

    The reset() method can be called to perform cleanUp and setUp automatically
    and potentially faster.
    """

    def addCleanup(
        self, cleanup: Callable[..., Any], *args: Any, **kwargs: Any
    ) -> None:
        """Add a clean function to be called from cleanUp.

        All cleanup functions are called - see cleanUp for details on how
        multiple exceptions are handled.

        If for some reason you need to cancel cleanups, call
        self._clear_cleanups.

        :param cleanup: A callable to call during cleanUp.
        :param args: Positional args for cleanup.
        :param kwargs: Keyword args for cleanup.
        :return: None
        """
        if self._cleanups is not None:
            self._cleanups.push(cleanup, *args, **kwargs)

    def addDetail(self, name: str, content_object: Any) -> None:
        """Add a detail to the Fixture.

        This may only be called after setUp has been called.

        :param name: The name for the detail being added. Overrides existing
            identically named details.
        :param content_object: The content object (meeting the
            testtools.content.Content protocol) being added.
        """
        if self._details is not None:
            self._details[name] = content_object

    def cleanUp(
        self, raise_first: bool = True
    ) -> Optional[
        List[
            Tuple[
                Optional[type[BaseException]],
                Optional[BaseException],
                Optional["TracebackType"],
            ]
        ]
    ]:
        """Cleanup the fixture.

        This function will free all resources managed by the Fixture, restoring
        it (and any external facilities such as databases, temporary
        directories and so forth_ to their original state.

        This should not typically be overridden, see addCleanup instead.

        cleanUp may be called once and only once after setUp() has been called.
        The base implementation of setUp will automatically call cleanUp if
        an exception occurs within setUp itself.

        :param raise_first: Deprecated parameter from before testtools gained
            MultipleExceptions. raise_first defaults to True. When True
            if a single exception is raised, it is reraised after all the
            cleanUps have run. If multiple exceptions are raised, they are
            all wrapped into a MultipleExceptions object, and that is reraised.
            Thus, to catch a specific exception from cleanUp, you need to catch
            both the exception and MultipleExceptions, and then check within
            a MultipleExceptions instance for the type you're catching.
        :return: A list of the exc_info() for each exception that occurred if
            raise_first was False
        """
        try:
            return self._cleanups(raise_errors=raise_first)  # type: ignore[misc]
        finally:
            self._remove_state()

    def _clear_cleanups(self) -> None:
        """Clean the cleanup queue without running them.

        This is a helper that can be useful for subclasses which define
        reset(): they may perform something equivalent to a typical cleanUp
        without actually calling the cleanups.

        This also clears the details dict.
        """
        self._cleanups = CallMany()
        self._details = {}
        self._detail_sources = []

    def _remove_state(self) -> None:
        """Remove the internal state.

        Called from cleanUp to put the fixture back into a not-ready state.
        """
        self._cleanups = None
        self._details = None
        self._detail_sources = None

    def __enter__(self) -> "Fixture":
        self.setUp()
        return self

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> Literal[False]:
        try:
            if self._cleanups is not None:
                self._cleanups()
        finally:
            self._remove_state()
        return False  # propagate exceptions from the with body.

    def getDetails(self) -> Dict[str, Any]:
        """Get the current details registered with the fixture.

        This does not return the internal dictionary: mutating it will have no
        effect. If you need to mutate it, just do so directly.

        :return: Dict from name -> content_object.
        """
        result = dict(self._details)  # type: ignore[arg-type]
        if self._detail_sources is not None:
            for source in self._detail_sources:
                combine_details(source.getDetails(), result)
        return result

    def setUp(self) -> None:
        """Prepare the Fixture for use.

        This should not be overridden. Concrete fixtures should implement
        _setUp. Overriding of setUp is still supported, just not recommended.

        After setUp has completed, the fixture will have one or more attributes
        which can be used (these depend totally on the concrete subclass).

        :raises: MultipleExceptions if _setUp fails. The last exception
            captured within the MultipleExceptions will be a SetupError
            exception.
        :return: None.

        :changed in 1.3: The recommendation to override setUp has been
            reversed - before 1.3, setUp() should be overridden, now it should
            not be.
        :changed in 1.3.1: BaseException is now caught, and only subclasses of
            Exception are wrapped in MultipleExceptions.
        """
        self._clear_cleanups()
        try:
            self._setUp()
        except BaseException:
            err: Tuple[
                Optional[type[BaseException]],
                Optional[BaseException],
                Optional["TracebackType"],
            ] = sys.exc_info()
            details: Dict[str, Any] = {}
            if gather_details is not None:
                # Materialise all details since we're about to cleanup.
                gather_details(self.getDetails(), details)
            else:
                details = self.getDetails()
            cleanup_errors = self.cleanUp(raise_first=False)
            errors = [err] + (cleanup_errors if cleanup_errors is not None else [])
            try:
                raise SetupError(details)
            except SetupError:
                errors.append(sys.exc_info())
            if err[0] is not None and issubclass(err[0], Exception):
                raise MultipleExceptions(*errors)
            else:
                if err[1] is not None:
                    raise err[1].with_traceback(err[2])

    def _setUp(self) -> None:
        """Template method for subclasses to override.

        Override this to customise the fixture. When overriding
        be sure to include self.addCleanup calls to restore the fixture to
        an un-setUp state, so that a single Fixture instance can be reused.

        Fixtures will never have a body in _setUp - calling super() is
        entirely at the discretion of subclasses.

        :return: None.
        """

    def reset(self) -> None:
        """Reset a setUp Fixture to the 'just setUp' state again.

        The default implementation calls
        self.cleanUp()
        self.setUp()

        but this function may be overridden to provide an optimised routine to
        achieve the same result.

        :return: None.
        """
        self.cleanUp()
        self.setUp()

    def useFixture(self, fixture: T) -> T:
        """Use another fixture.

        The fixture will be setUp, and self.addCleanup(fixture.cleanUp) called.
        If the fixture fails to set up, useFixture will attempt to gather its
        details into this fixture's details to aid in debugging.

        :param fixture: The fixture to use.
        :return: The fixture, after setting it up and scheduling a cleanup for
           it.
        :raises: Any errors raised by the fixture's setUp method.
        """
        try:
            fixture.setUp()
        except MultipleExceptions as e:
            if e.args[-1][0] is SetupError:
                if self._details is not None:
                    combine_details(e.args[-1][1].args[0], self._details)
            raise
        except:
            # The child failed to come up and didn't raise MultipleExceptions
            # which we can understand... capture any details it has (copying
            # the content, it may go away anytime).
            if gather_details is not None:
                if self._details is not None:
                    gather_details(fixture.getDetails(), self._details)
            raise
        else:
            self.addCleanup(fixture.cleanUp)
            # Calls to getDetails while this fixture is setup will return
            # details from the child fixture.
            if self._detail_sources is not None:
                self._detail_sources.append(fixture)
            return fixture


class FunctionFixture(Fixture):
    """An adapter to use function(s) as a Fixture.

    Typically used when an existing object or function interface exists but you
    wish to use it as a Fixture (e.g. because fixtures are in use in your test
    suite and this will fit in better).

    To adapt an object with differently named setUp and cleanUp methods:
    fixture = FunctionFixture(object.install, object.__class__.remove)
    Note that the indirection via __class__ is to get an unbound method
    which can accept the result from install. See also MethodFixture which
    is specialised for objects.

    To adapt functions:
    fixture = FunctionFixture(tempfile.mkdtemp, shutil.rmtree)

    With a reset function:
    fixture = FunctionFixture(setup, cleanup, reset)

    :ivar fn_result: The result of the setup_fn. Undefined outside of the
        setUp, cleanUp context.
    """

    def __init__(
        self,
        setup_fn: Callable[[], Any],
        cleanup_fn: Optional[Callable[[Any], None]] = None,
        reset_fn: Optional[Callable[[Any], Any]] = None,
    ) -> None:
        """Create a FunctionFixture.

        :param setup_fn: A callable which takes no parameters and returns the
            thing you want to use. e.g.
            def setup_fn():
                return 42
            The result of setup_fn is assigned to the fn_result attribute bu
            FunctionFixture.setUp.
        :param cleanup_fn: Optional callable which takes a single parameter,
            which must be that which is returned from the setup_fn. This is
            called from cleanUp.
        :param reset_fn: Optional callable which takes a single parameter like
            cleanup_fn, but also returns a new object for use as the fn_result:
            if defined this replaces the use of cleanup_fn and setup_fn when
            reset() is called.
        """
        super(FunctionFixture, self).__init__()
        self.setup_fn = setup_fn
        self.cleanup_fn = cleanup_fn
        self.reset_fn = reset_fn

    def _setUp(self) -> None:
        fn_result = self.setup_fn()
        self._maybe_cleanup(fn_result)

    def reset(self) -> None:
        if self.reset_fn is None:
            super(FunctionFixture, self).reset()
        else:
            self._clear_cleanups()
            fn_result = self.reset_fn(self.fn_result)
            self._maybe_cleanup(fn_result)

    def _maybe_cleanup(self, fn_result: Any) -> None:
        self.addCleanup(delattr, self, "fn_result")
        if self.cleanup_fn is not None:
            self.addCleanup(self.cleanup_fn, fn_result)
        self.fn_result = fn_result


class MethodFixture(Fixture):
    """An adapter to use a function as a Fixture.

    Typically used when an existing object exists but you wish to use it as a
    Fixture (e.g. because fixtures are in use in your test suite and this will
    fit in better).

    To adapt an object with setUp / tearDown methods:
    fixture = MethodFixture(object)
    If setUp / tearDown / reset are missing, they simply won't be called.

    The object is exposed on fixture.obj.

    To adapt an object with differently named setUp and cleanUp methods:
    fixture = MethodFixture(object, setup=object.mySetUp,
        teardown=object.myTearDown)

    With a differently named reset function:
    fixture = MethodFixture(object, reset=object.myReset)

    :ivar obj: The object which is being wrapped.
    """

    def __init__(
        self,
        obj: Any,
        setup: Optional[Callable[[], None]] = None,
        cleanup: Optional[Callable[[], None]] = None,
        reset: Optional[Callable[[], None]] = None,
    ) -> None:
        """Create a MethodFixture.

        :param obj: The object to wrap. Exposed as fixture.obj
        :param setup: A method which takes no parameters. e.g.
            def setUp(self):
                self.value = 42
            If setup is not supplied, and the object has a setUp method, that
            method is used, otherwise nothing will happen during fixture.setUp.
        :param cleanup: Optional method to cleanup the object's state. If
            not supplied the method 'tearDown' is used if it exists.
        :param reset: Optional method to reset the wrapped object for use.
            If not supplied, then the method 'reset' is used if it exists,
            otherwise cleanUp and setUp are called as per Fixture.reset().
        """
        super(MethodFixture, self).__init__()
        self.obj = obj
        if setup is None:
            setup = getattr(obj, "setUp", None)
            if setup is None:

                def setup() -> None:
                    return None

        self._setup = setup
        if cleanup is None:
            cleanup = getattr(obj, "tearDown", None)
            if cleanup is None:

                def cleanup() -> None:
                    return None

        self._cleanup = cleanup
        if reset is None:
            reset = getattr(obj, "reset", None)
        self._reset = reset

    def _setUp(self) -> None:
        self._setup()

    def cleanUp(
        self, raise_first: bool = True
    ) -> Optional[
        List[
            Tuple[
                Optional[type[BaseException]],
                Optional[BaseException],
                Optional["TracebackType"],
            ]
        ]
    ]:
        result = super(MethodFixture, self).cleanUp(raise_first)
        self._cleanup()
        return result

    def reset(self) -> None:
        if self._reset is None:
            super(MethodFixture, self).reset()
        else:
            self._reset()


class CompoundFixture(Fixture):
    """A fixture that combines many fixtures.

    :ivar fixtures: The list of fixtures that make up this one. (read only).
    """

    def __init__(self, fixtures: Iterable["Fixture"]) -> None:
        """Construct a fixture made of many fixtures.

        :param fixtures: An iterable of fixtures.
        """
        super(CompoundFixture, self).__init__()
        self.fixtures = list(fixtures)

    def _setUp(self) -> None:
        for fixture in self.fixtures:
            self.useFixture(fixture)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "WarningsCapture",
    "WarningsFilter",
]

import warnings
from typing import Any, Dict, List, Optional, cast

import fixtures
from fixtures._fixtures.monkeypatch import MonkeyPatch


class WarningsCapture(fixtures.Fixture):
    """Capture warnings.

    While ``WarningsCapture`` is active, warnings will be captured by
    the fixture (so that they can be later analyzed).

    :attribute captures: A list of warning capture ``WarningMessage`` objects.
    """

    captures: List[warnings.WarningMessage]

    def _showwarning(self, *args: Any, **kwargs: Any) -> None:
        self.captures.append(warnings.WarningMessage(*args, **kwargs))

    def _setUp(self) -> None:
        patch = MonkeyPatch("warnings.showwarning", self._showwarning)
        self.useFixture(patch)
        self.captures = []


class WarningsFilter(fixtures.Fixture):
    """Configure warnings filters.

    While ``WarningsFilter`` is active, warnings will be filtered per
    configuration.
    """

    def __init__(self, filters: Optional[List[Dict[str, Any]]] = None) -> None:
        """Create a WarningsFilter fixture.

        :param filters: An optional list of dictionaries with arguments
            corresponding to the arguments to
            :py:func:`warnings.filterwarnings`. For example::

                [
                    {
                        'action': 'ignore',
                        'message': 'foo',
                        'category': DeprecationWarning,
                    },
                ]

            Order is important: entries closer to the front of the list
            override entries later in the list, if both match a particular
            warning.

            Alternatively, you can configure warnings within the context of the
            fixture.

            See `the Python documentation`__ for more information.

        __: https://docs.python.org/3/library/warnings.html#the-warnings-filter
        """
        super().__init__()
        self.filters = filters or []

    def _setUp(self) -> None:
        self._original_warning_filters = list(warnings.filters)

        for filt in self.filters:
            warnings.filterwarnings(**filt)

        self.addCleanup(self._reset_warning_filters)

    def _reset_warning_filters(self) -> None:
        filters_list = cast(List[Any], warnings.filters)
        filters_list.clear()
        filters_list.extend(self._original_warning_filters)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2011, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "PackagePathEntry",
]

import sys

from fixtures import Fixture


class PackagePathEntry(Fixture):
    """Add a path to the path of a python package.

    The python package needs to be already imported.

    If this new path is already in the packages __path__ list then the __path__
    list will not be altered.
    """

    def __init__(self, packagename: str, directory: str) -> None:
        """Create a PackagePathEntry.

        :param directory: The directory to add to the package.__path__.
        """
        self.packagename = packagename
        self.directory = directory

    def _setUp(self) -> None:
        path = sys.modules[self.packagename].__path__
        if self.directory in path:
            return
        self.addCleanup(path.remove, self.directory)
        path.append(self.directory)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (c) 2012, Robert Collins <robertc@robertcollins.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.

__all__ = [
    "ByteStream",
    "DetailStream",
    "StringStream",
]

import io
from typing import Callable, Generic, IO, Tuple, TypeVar, Union

from fixtures import Fixture

# Type variable for the stream type
T = TypeVar("T", IO[bytes], IO[str])


class Stream(Generic[T], Fixture):
    """Expose a file-like object as a detail.

    :attr stream: The file-like object.
    """

    stream: T

    def __init__(
        self,
        detail_name: str,
        stream_factory: Callable[[], Tuple[T, Union[IO[bytes], IO[str]]]],
    ) -> None:
        """Create a ByteStream.

        :param detail_name: Use this as the name of the stream.
        :param stream_factory: Called to construct a pair of streams:
            (write_stream, content_stream).
        """
        self._detail_name = detail_name
        self._stream_factory: Callable[[], Tuple[T, Union[IO[bytes], IO[str]]]] = (
            stream_factory
        )

    def _setUp(self) -> None:
        # Available with the fixtures[streams] extra.
        from testtools.content import content_from_stream

        write_stream, read_stream = self._stream_factory()
        self.stream = write_stream
        self.addDetail(
            self._detail_name,
            content_from_stream(read_stream, seek_offset=0),
        )


def _byte_stream_factory() -> Tuple[IO[bytes], IO[bytes]]:
    result = io.BytesIO()
    return (result, result)


def ByteStream(detail_name: str) -> Stream[IO[bytes]]:
    """Provide a file-like object that accepts bytes and expose as a detail.

    :param detail_name: The name of the detail.
    :return: A fixture which has an attribute `stream` containing the file-like
        object.
    """
    return Stream(detail_name, _byte_stream_factory)


def _string_stream_factory() -> Tuple[IO[str], IO[bytes]]:
    lower = io.BytesIO()
    upper = io.TextIOWrapper(lower, encoding="utf8")
    # See http://bugs.python.org/issue7955
    upper._CHUNK_SIZE = 1  # type: ignore[attr-defined]
    return upper, lower


def StringStream(detail_name: str) -> Stream[IO[str]]:
    """Provide a file-like object that accepts strings and expose as a detail.

    :param detail_name: The name of the detail.
    :return: A fixture which has an attribute `stream` containing the file-like
        object.
    """
    return Stream(detail_name, _string_stream_factory)


def DetailStream(detail_name: str) -> Stream[IO[bytes]]:
    """Deprecated alias for ByteStream."""
    return ByteStream(detail_name)



#  fixtures: Fixtures with cleanups for testing and convenience.
#
# Copyright (C) 2011, Martin Pool <mbp@sourcefrog.net>
#
# Licensed under either the Apache License, Version 2.0 or the BSD 3-clause
# license at the users choice. A copy of both licenses are available in the
# project source as Apache-2.0 and BSD. You may not use this file except in
# compliance with one of these two licences.
#
# Unless required by applicable law or agreed to in writing, software
# distributed under these licenses is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the
# license you chose for the specific language governing permissions and
# limitations under that license.


"""Timeout fixture."""

import signal
from typing import Any, Callable, Optional

import fixtures

__all__ = [
    "Timeout",
    "TimeoutException",
]


class TimeoutException(Exception):
    """Timeout expired"""


class Timeout(fixtures.Fixture):
    """Fixture that aborts the contained code after a number of seconds.

    The interrupt can be either gentle, in which case TimeoutException is
    raised, or not gentle, in which case the process will typically be aborted
    by SIGALRM.

    Cautions:
     * This has no effect on Windows.
     * Only one Timeout can be used at any time per process.
    """

    timeout_secs: int
    alarm_fn: Optional[Callable[[int], int]]
    gentle: bool

    def __init__(self, timeout_secs: int, gentle: bool) -> None:
        self.timeout_secs = timeout_secs
        self.alarm_fn = getattr(signal, "alarm", None)
        self.gentle = gentle

    def signal_handler(self, signum: int, frame: Any) -> None:
        raise TimeoutException()

    def _setUp(self) -> None:
        if self.alarm_fn is None:
            return  # Can't run on Windows
        if self.gentle:
            # Install a handler for SIGARLM so we can raise an exception rather
            # than the default handler executing, which kills the process.
            old_handler = signal.signal(signal.SIGALRM, self.signal_handler)
        # We add the slarm cleanup before the cleanup for the signal handler,
        # otherwise there is a race condition where the signal handler is
        # cleaned up but the alarm still fires.
        if self.alarm_fn is not None:
            self.addCleanup(lambda: self.alarm_fn(0) if self.alarm_fn else None)
            self.alarm_fn(self.timeout_secs)
        if self.gentle:
            self.addCleanup(lambda: signal.signal(signal.SIGALRM, old_handler))
